/* wptGPGParser.cpp - Config file parser
 *	Copyright (C) 2002, 2003, 2004, 2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <windows.h>

#include "wptTypes.h"
#include "wptW32API.h"
#include "wptGPG.h"
#include "wptErrors.h"


/* Add the option @n to the linked list @list. */
static void
add_option (conf_option_t *list, conf_option_t n)
{
    conf_option_t t;

    if (!*list)
	*list = n;
    else {
	for (t = *list; t->next; t = t->next)
	    ;
	t->next = n;
    }
}

/* Allocate a new option and return it. */
static conf_option_t
new_option (void)
{
    conf_option_t e;

    e = new conf_option_s;
    if (!e)
	BUG (0);
    memset (e, 0, sizeof *e);
    return e;
}


static void
add_opaque_option (config_file_t opt, const char *line, int used)
{
    conf_option_t e;

    e = new_option ();
    e->used = used > 0? 1 : 0;
    e->name = m_strdup (line);
    e->type = ENTRY_OPAQUE;
    add_option (&opt->list, e);
}


static void
add_single_option (config_file_t opt, const char *name, int used)
{
    conf_option_t e;

    e = new_option ();
    e->used = used > 0? 1 : 0;
    e->name = m_strdup (name);
    e->type = ENTRY_SINGLE;
    add_option (&opt->list, e);
}


static int
add_group_option (config_file_t opt, const char *name, int used)
{
    conf_option_t e;

    e = new_option ();
    e->used = used > 0? 1 : 0;
    e->name = m_strdup (name);
    e->type = ENTRY_GROUP;
    add_option (&opt->list, e);
    return 0;
}


static size_t
count_whitespaces (char *line)
{
    size_t i;
    size_t ncount = 0;

    if (!strchr (line, ' '))
	return 0;
    for( i = 0; i < strlen (line); i++) {
	if (line[i] == ' ')
	    ncount++;
    }
    return ncount;
}


static int
add_multi_option (config_file_t opt, char *line, int used)
{
    conf_option_t e;
    char *p;
    int state = 0;

    e = new_option ();
    if (count_whitespaces (line) == 1) {
	while ((p = strsep (&line, " ")) && state != 2) {
	    switch (state) {
	    case 0: e->name = m_strdup (p); break;
	    case 1: e->val = m_strdup (p);  break;
	    }
	    state++;
	}
    }
    else {
	p = strchr (line, ' ');
	if (!p)
	    return -1;
	state = p - line;
	e->name = new char[state + 1];
	if (!e->name)
	    BUG (NULL);
	memcpy (e->name, line, state);
	e->name[state] = '\0';
	strncpy (e->name, line, state);
	e->val = m_strdup (line + state + 1);
    }
    e->used = used;
    e->type = ENTRY_MULTI;
    add_option (&opt->list, e);

    return 0;
}


static conf_group_t
new_group (conf_group_t *grp, const char *name)
{
    conf_group_t g, t;

    g = new conf_group_s;    
    if (!g)
	BUG (NULL);
    memset (g, 0, sizeof *g);
    g->used = 1;
    g->name = m_strdup (name);
    if (!*grp)
	*grp = g;
    else {
	for (t = *grp; t->next; t = t->next)
	    ;
	t->next = g;
    }
    return g;
}


static const char*
group_from_cfgfile (config_file_t opt, const char *line)
{
    char *p, *buf, *buf2, *name = NULL;
    conf_group_t g;
    conf_member_t m = NULL, t;

    p = strchr (line, '=');
    if (!p || strncmp (line, "group ", 6))
        return NULL;
    buf = m_strdup (line + (p - line + 1));
    line += 6;
    p = strchr (line, '=');
    if (p) {
        size_t pos = p - line;
	name = new char[pos + 1];
	if (!name)
	    BUG( NULL );
	memset (name, 0, pos+1);
        strncpy (name, line, pos);
    }
    g = new_group (&opt->grp, name);
    buf2 = buf;
    while ((p = strsep (&buf2, " ")) && *p) {
	m = new conf_member_s;
        if (!m)
            BUG (0);
	memset (m, 0, sizeof *m);
        m->used = 1;
        m->name = m_strdup (p);
	if (!g->list)
	    g->list = m;
        else {
	    for (t=g->list; t->next; t = t->next)
		;
	    t->next = m;
	}
    }
    free_if_alloc (buf);
    if (!m)
        return NULL;
    return g->name;
}


void
conf_release_group (conf_group_t grp)
{
    conf_member_t m, m2;
    
    m = grp->list;
    while (m) {
        m2 = m->next;
        free_if_alloc (m->name);
        free_if_alloc (m);
        m = m2;
    }
}


static void
release_option (conf_option_t opt)
{
    if (!opt)
	return;
    free_if_alloc (opt->name);
    free_if_alloc (opt->val);
    free_if_alloc (opt);
}


void
release_config (config_file_t opt)
{
    conf_option_t e, e2;
    conf_group_t g, g2;

    if (!opt)
	return;
    e = opt->list;
    while (e) {
        e2 = e->next;
        release_option (e);
        e = e2;
    }
    g = opt->grp;
    while (g) {
        g2 = g->next;
        conf_release_group (g);
        g = g2;
    }
    free_if_alloc (opt);
}


int
commit_config (const char *file, config_file_t opt)
{
    FILE *inp;
    conf_option_t e;
    conf_group_t g;
    conf_member_t m;

    inp = fopen (file, "w+b");
    if (!inp)
        return WPTERR_FILE_OPEN;
    for (e = opt->list; e; e = e->next) {
        if (!e->used)
            continue;
        switch (e->type) {
        case ENTRY_OPAQUE:
            fprintf (inp, "%s", e->name);
            break;
            
        case ENTRY_MULTI:
            fprintf (inp, "%s %s\r\n", e->name, e->val);
            break;

        case ENTRY_SINGLE:
            fprintf (inp, "%s\r\n", e->name);
            break;

        case ENTRY_GROUP:
            g = conf_find_group (opt, e->name);
            if (g && g->used) {
                fprintf (inp, "group %s=", g->name);
                for (m = g->list; m; m = m->next) {
                    if (m->used)
                        fprintf (inp, "%s ", m->name);
                }
                fprintf (inp, "\r\n");
            }
            break;
        }
    }
    fclose (inp);
    return 0;
}


void 
new_config (config_file_t *r_opt)
{
    config_file_t opt;

    opt = new conf_file_s;
    if (!opt)
	BUG (NULL);
    memset (opt, 0, sizeof *opt);
    *r_opt = opt;
}


int
parse_config (const char *file, config_file_t *r_opt)
{
    FILE *inp;
    char buf[1024], *p;
    config_file_t opt = NULL;

    if (!r_opt)
	return -1;

    inp = fopen( file, "rb");
    if (!inp) {
	inp = fopen (file, "wb");
	if (inp == NULL)
	    return -1;
    }
    opt = new conf_file_s;
    if (!opt)
	BUG (NULL);
    memset (opt, 0, sizeof *opt);
    while (!feof (inp))  {
        p = fgets (buf, DIM (buf) -1, inp);
        if (!p)
            break;
        if (*p == '#' || *p == '\r' || *p == '\n' || *p == '\t') {
            add_opaque_option (opt, p, 1);
            continue;
        }
	if( strstr (p, "\r\n" ))
	    p[strlen (p) - 2] = '\0';
	else if (strstr (p, "\n"))
	    p[strlen (p) - 1] = '\0';
        if (!strchr( p, ' ' ))
            add_single_option (opt, p, 1);
        else if (!strncmp (p, "group", 5)) {
            const char *s = group_from_cfgfile (opt, p);
            if (s)
                add_group_option (opt, s, 1);
        }
        else
            add_multi_option (opt, p, 1);
    }
    fclose (inp);
    *r_opt = opt;
    return 0;
}


/* Search for an option with the name @str and
   return this option if it exists.
   Return value: NULL if not found. */
conf_option_t
conf_find_option (config_file_t opt, const char *str)
{
    conf_option_t e;

    for (e = opt->list; e; e = e->next) {
        if (e->type == ENTRY_OPAQUE)
            continue;
        if (!stricmp (e->name, str))
            break;
    }

    return e;
}


conf_group_t
conf_find_group (config_file_t opt, const char *str)
{
    conf_group_t g;

    for (g = opt->grp; g; g = g->next) {
        if (!stricmp( g->name, str))
            break;
    }

    return g;
}


conf_member_t
conf_find_member (config_file_t opt, const char *grp, const char *str)
{
    conf_member_t m;
    conf_group_t g = conf_find_group (opt, grp);

    if (!g)
	return NULL;
        
    for (m = g->list; m; m = m->next) {
        if (!stricmp (m->name, str))
	    return m;
    }
    return NULL;
}


int
conf_delete_group (config_file_t opt, const char *str)
{
    conf_group_t g = conf_find_group (opt, str);

    if (g)
        g->used = 0;
    return g? 0 : -1;
}


int
conf_delete_member (config_file_t opt, const char *grp, const char *str)
{
    conf_member_t m = conf_find_member (opt, grp, str);

    if (m)
	m->used = 0;
    return m? 0 : -1;
}

    
int
conf_delete_option (config_file_t opt, const char * str)
{
    conf_option_t e = conf_find_option (opt, str);

    if (e)
        e->used = 0;
    return e? 0 : -1;
}


int
conf_add_entry_int (config_file_t opt, int type, const char *name, int val)
{
    char buf[64];

    _snprintf (buf, DIM (buf)-1, "%d", val);
    conf_add_entry (opt, type, name, buf);
    return 0;
}

int
conf_add_entry (config_file_t opt, int type, const char *name, const char *val)
{
    conf_option_t e;

    e = new_option ();
    e->used = 1;
    e->type = type;
    e->name = m_strdup (name);
    e->val = val? m_strdup (val) : NULL;
    add_option (&opt->list, e);
    return 0;
}


/* Modify the value of the entry with the name @name.
   If the entry were not found, add it to the option list. */
int
conf_modify_entry (config_file_t opt, int type, const char *name, const char *val)
{
    conf_option_t e;
    int rc = 0;

    e = conf_find_option (opt, name);
    if (!e)
	rc = conf_add_entry (opt, type, name, val);
    else if (type != ENTRY_SINGLE && val != NULL) {
	free_if_alloc (e->val);
	e->val = m_strdup (val);
	rc = 0;
    }

    return rc;
}

    
int
conf_add_member (config_file_t opt, const char *grp, const char *str)
{
    conf_group_t g = conf_find_group (opt, grp);
    conf_member_t m, t;

    if (!g)
	return -1;
    
    m = new conf_member_s;    
    if (!m)    
        BUG (NULL);
    memset (m, 0, sizeof *m);
    m->used = 1;
    m->name = m_strdup (str);
    if( !g->list )
	g->list = m;
    else {
	for( t = g->list; t->next; t = t->next )
	    ;
	t->next = m;	
    }   
    return 0;
}


int
conf_add_group (config_file_t opt, const char *str)
{
    conf_group_t g, t;

    g = new conf_group_s;
    if (!g)
	BUG (NULL);
    memset (g, 0, sizeof *g);
    g->used = 1;
    g->name = m_strdup (str);
    
    if (!opt->grp)
	opt->grp = g;
    else {
	for (t = opt->grp; t->next; t = t->next)
	    ;
	t->next = g;
    }
    conf_add_entry (opt, ENTRY_GROUP, str, NULL);
    return 0;
}
