/* wptHotkey.cpp - Systemwide hotkeys
 *	Copyright (C) 2001-2003, 2006, 2008, 2009 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>

#include "wptHotkey.h"
#include "wptGPG.h"
#include "wptRegistry.h"
#include "wptErrors.h"
#include "wptNLS.h"
#include "wptTypes.h"


/* List of all predefined hotkeys. */
hotkey_s wpt_hotkeys[]  = {
    /* ALT+CTRL+ */
    {"ClipEncrypt", 1, WPT_CLIP_ENCRYPT_ID,           1, 0, 0x45}, /* E */
    {"ClipDecrypt", 1, WPT_CLIP_DECRYPT_VERIFY_ID,    1, 0, 0x44}, /* D */
    {"ClipSign",    1, WPT_CLIP_SIGN_ID,	      1, 0, 0x53}, /* S */
    {"ClipSignEnc", 1, WPT_CLIP_SIGN_ENCRYPT_ID,      1, 0, 0x42}, /* B */
    /* ALT+SHIFT+ */
    {"CwsEncrypt",  1, WPT_CURRWND_ENCRYPT_ID,        0, 1, 0x45}, /* E */
    {"CwsDecrypt",  1, WPT_CURRWND_DECRYPT_VERIFY_ID, 0, 1, 0x44}, /* D */
    {"CwsSign",     1, WPT_CURRWND_SIGN_ID,           0, 1, 0x53}, /* S */
    {"CwsSignEnc",  1, WPT_CURRWND_SIGN_ENCRYPT_ID,   0, 1, 0x42}, /* B */
    /* Agent ALT+CTRL+ */
    {"AgentForget", 1, WPT_AGENT_FORGET_ID,           1, 0, 0x46}, /* F */
    {0}
};


/* Global error variable. */
static int winpt_hotkey_err = 0;

const char*
hotkeys_strerror (void)
{
    switch (winpt_hotkey_err) {
    case 0: return NULL;
    case 1: return _("Clipboard Encrypt (ALT+CTRL+E)");
    case 2: return _("Clipboard Decrypt/Verify (ALT+CTRL+D)");
    case 3: return _("Clipboard Sign (ALT+CTRL+S)");
    case 4: return _("Clipboard Sign Encrypt (ALT+CTRL+B)");
    case 5: return _("Current Window Encrypt (ALT+SHIFT+E)");
    case 6: return _("Current Window Decrypt/Verify (ALT+SHIFT+D)");
    case 7: return _("Current Window Sign (ALT+SHIFT+S)");
    case 8: return _("Current Window Sign Encrypt (ALT+SHIFT+B)");
    default:return _("Unknown Hotkey");
    }
    return _("Unknown Hotkey");
}


/* Register all enabled hotkeys. */
int 
hotkeys_register (HWND wnd)
{
    int rc = 0;
	
    winpt_hotkey_err = 0; /* reset */
    for (int j=0; wpt_hotkeys[j].id != 0; j++) {
	if (wpt_hotkeys[j].enabled) {
	    rc = hotkey_register_single (wnd, &wpt_hotkeys[j]);
	    if (rc) {
		winpt_hotkey_err = j+1;
		break;	
	    }
	}	
    }

    return rc;
}


/* Unregister all hotkeys. */
int 
hotkeys_unregister (HWND wnd)
{
    for (int j=0; wpt_hotkeys[j].id != 0; j++)
	hotkey_unregister_single (wnd, &wpt_hotkeys[j]);

    return 0;
}


/* Register a single hotkey @hk. */
int
hotkey_register_single (HWND wnd, hotkey_t hk)
{
    int rc;
    int keymod = 0;

    if (!hk->enabled)
	return 0;
    if (hk->alt_ctrl)
	keymod = MOD_CONTROL | MOD_ALT;
    else if (hk->alt_shift)
	keymod = MOD_ALT | MOD_SHIFT;

    rc = RegisterHotKey (wnd, hk->id, keymod, hk->key);
    return rc == 0? WPTERR_HOTKEY : 0;
}


/* Unregister a single hotkey @hk. */
int
hotkey_unregister_single (HWND wnd, hotkey_t hk)
{
    int rc;

    rc = UnregisterHotKey (wnd, hk->id);
    return rc==0? WPTERR_HOTKEY: 0;
}


/* Disable the given hotkey in @hk. */
void
hotkey_disable (hotkey_t hk)
{
    hk->enabled = 0;
}


/* Enable the given hotkey in @hk.
   Set the hotkey sequence to @key. */
void
hotkey_enable (hotkey_t hk, const char *key)
{
    if (strlen (key) > 2)
	BUG (0);
    hk->key = *key;
    hk->enabled = 1;
}
