/* wptKeyRevokeDlg.cpp - Key revocation dialog
 *	Copyright (C) 2001-2003, 2005-2006, 2009 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>

#include "resource.h"
#include "gpgme.h"
#include "wptErrors.h"
#include "wptGPG.h"
#include "wptW32API.h"
#include "wptTypes.h"
#include "wptCommonCtl.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptDlgs.h"
#include "wptNLS.h"
#include "wptUTF8.h"

void secure_filename (char *file, size_t len);


/* Generate a file template for the cert based on the key 
   with the keyid @keyid. */
static void
mk_cert_fname (const char *keyid, char *fname, size_t flen)
{
    winpt_key_s k;
    const char *filename = _("Revocation-Certificate");

    memset (&k, 0, sizeof (k));
    if (winpt_get_pubkey (keyid, &k))
	BUG (NULL);
    _snprintf (fname, flen-1, "%s_%s.asc", k.ext->uids->name, filename);
    secure_filename (fname, strlen (fname));
}


/* Release the cert data and store it in the file @fname. */
static void
release_cert_as_file (char *revcert, const char *fname)
{
    gpgme_error_t err;
    gpgme_data_t rev;

    err = gpgme_data_new_from_mem (&rev, revcert, strlen (revcert), 1);
    if (!err)
	err = gpg_data_release_and_set_file (rev, fname);
    if (err)
	msg_box (NULL, gpgme_strerror (err), _("Key Revocation Cert"), MB_ERR);
    safe_free (revcert);
}


static void
on_init_dialog (HWND dlg)
{
    HWND list;

    SetWindowText (dlg, _("Key Revocation Cert"));
    SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
    SetDlgItemText (dlg, IDC_KEYREVOKE_REASONINF, _("Reason for revocation"));
    SetDlgItemText (dlg, IDC_KEYREVOKE_OPTINF, _("Optional description text"));
    SetDlgItemText (dlg, IDC_KEYREVOKE_PWDINF, _("&Passphrase:"));
    SetDlgItemText (dlg, IDC_KEYREVOKE_OUTINF, _("Output file:"));
    list = GetDlgItem (dlg, IDC_KEYREVOKE_REASON);
    listbox_add_string (list, _("0. No reason specified"));
    listbox_add_string (list, _("1. Key has been compromised"));
    listbox_add_string (list, _("2. Key is superseded"));
    listbox_add_string (list, _("3. Key is no longer used"));
    SendMessage (list, LB_SETCURSEL, (WPARAM)0, 0);
    SetForegroundWindow (dlg);
    center_window (dlg, NULL);
}

/* Generate the data expected by the gpg command handler. */
static char* 
generate_revoke_input (int code, const char *cmt, const char *pass)
{
    const char *fmt;
    char *p;
    size_t n;

    fmt = "Y\n"	    /*  gen_revoke.okay */
	  "%d\n"    /* ask_revocation_reason.code */
	  "%s\n"    /* ask_revocation_reason.text */
	  "%s"	    /* text != NULL '\n' otherwise '' */
	  "Y\n"	    /* ask_revocation_reason.okay */
	  "%s\n";   /* passphrase.enter. */
    n = strlen (fmt) + 32;
    if (pass)
	n += strlen (pass) + 1;
    if (cmt)
	n += strlen (cmt) + 1;
    p = new char[n+1];
    if (!p)
	BUG (0);
    sprintf (p, fmt, code, cmt? cmt : "", cmt? "\n" : "", pass? pass : "");
    return p;
}


/* Dialog box procedure for key revocation. */
BOOL CALLBACK
key_revoke_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{	
    static winpt_key_t k;
    gpgme_error_t err;
    HWND list;
    int idx, use_desc;
    char file[256];
    char *pwd=NULL;
    char *desc=NULL;
    char *inp_data = NULL, *revcert=NULL;
    const char *warning =
	_("Please move this certificate to a medium where it can be "
          "stored in a safe place (floppy, CDR, etc..).\n"
          "If an attacker gets access to this certificate he can use it to "
          "render your key unusable!");
    
    switch( msg ) {
    case WM_INITDIALOG:
        if (!lparam)
	    BUG (0);
        k = (winpt_key_t)lparam;
        on_init_dialog (dlg);
        return TRUE;
        
    case WM_COMMAND:
        switch (LOWORD (wparam)) {
	case IDC_KEYREVOKE_CHOOSE:
	    const char *s;
	    mk_cert_fname (k->keyid, file, sizeof (file)-1);
	    s = get_filesave_dlg (dlg, _("Choose File to save the Certificate"), NULL, file);
	    if (s && *s)
		SetDlgItemText (dlg, IDC_KEYREVOKE_FILE, s);
	    return TRUE;

        case IDOK:
            list = GetDlgItem (dlg, IDC_KEYREVOKE_REASON);
            idx = SendMessage (list, LB_GETCURSEL, 0, 0);
            if (idx < 0 || idx > 3) {
		msg_box (dlg, _("Please select a reason."),
			 _("Key Revocation Cert"), MB_ERR);
                return TRUE;
            }
            if (!GetDlgItemText (dlg, IDC_KEYREVOKE_FILE, file, 
				 DIM (file)-1)) {
		msg_box (dlg, _("Please enter a file name."),
			 _("Key Revocation Cert"), MB_ERR);
                return TRUE;
            }
	    if (check_file_name (file, IS_PATH)) {
		msg_box (dlg, _("The file name contains one or more illegal characters."),
			 _("Key Revocation Cert"), MB_ERR);
		return TRUE;
	    }

            use_desc = 1;
            if (!GetDlgItemText_utf8 (dlg, IDC_KEYREVOKE_TEXT, &desc))
                use_desc = 0;
            if (!GetDlgItemText_utf8 (dlg, IDC_KEYREVOKE_PWD, &pwd)) {
		msg_box (dlg, _("Please enter the passphrase."), 
			 _("Key Revocation Cert"), MB_ERR);
                return TRUE;
            }
          
	    inp_data = generate_revoke_input (idx, desc, pwd);
            err = gpg_revoke_cert (k->internal, inp_data, k->keyid, &revcert);
	    sfree_if_alloc (inp_data);
	    sfree_if_alloc (desc);
	    sfree_if_alloc (pwd);
            if (err) {
                msg_box (dlg, gpgme_strerror (err), _("Key Revocation Cert"), MB_ERR);
		safe_free (revcert);
		return TRUE;
	    }
	    else {
		show_msg (dlg, 1000, _("Revocation certificate generated."));
		msg_box (dlg, warning, _("Key Revocation Cert"), MB_INFO);
		release_cert_as_file (revcert, file);
		EndDialog (dlg, TRUE);
	    }
            return TRUE;
            
        case IDCANCEL:
            EndDialog (dlg, FALSE);
            return TRUE;
        }
        break;
    }
    
    return FALSE;
}
