/* wptListView.cpp - Dynamic list view control
 *	Copyright (C) 2000-2007, 2009 Timo Schulz
 *	Copyright (C) 2004 Andreas Jobs
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <windows.h>
#include <commctrl.h>

#include "wptCommonCtl.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptGPG.h"
#include "wptKeylist.h"
#include "wptUTF8.h"
#include "resource.h"


void
listview_new (listview_ctrl_t *ctx, HWND ctrl)
{
    struct listview_ctrl_s *c;

    c = new struct listview_ctrl_s;
    if (!c)
	BUG (NULL);
    memset (c, 0, sizeof *c);
    c->ctrl = ctrl? ctrl : NULL;
    c->hil = NULL;
    *ctx = c;
}


void
listview_release (listview_ctrl_t ctx)
{
    if (!ctx)
	return;

    ctx->cols = 0;
    ctx->ctrl = NULL;
    ctx->items = 0;
    if (ctx->hil)
	ImageList_Destroy (ctx->hil);
    delete ctx;
}


int
listview_set_column_width (listview_ctrl_t ctx, int col, int width)
{
    LVCOLUMN lvc;

    memset (&lvc, 0, sizeof (lvc));
    lvc.mask = LVCF_WIDTH;
    lvc.cx = width;
    ListView_SetColumn (ctx->ctrl, col, &lvc);
    return 0;
}


int
listview_add_column (listview_ctrl_t ctx, listview_column_t col)
{	
    int rc = 0;
    LV_COLUMN lvc;

    memset (&lvc, 0, sizeof lvc);
    lvc.mask = LVCF_TEXT | LVCF_WIDTH | LVCF_SUBITEM;
    lvc.pszText = col->fieldname;
    lvc.cx = col->width;
    lvc.fmt = LVCFMT_LEFT;
    lvc.iSubItem = col->pos;
    if (ListView_InsertColumn (ctx->ctrl, col->pos, &lvc) == -1)
	rc = 1;
    ctx->cols++;
    return rc;
}


int 
listview_add_item_pos (listview_ctrl_t ctx, int pos)
{
    int rc = 0;
    LV_ITEM lvi;

    memset (&lvi, 0, sizeof lvi);
    lvi.iItem = pos;
    if (ListView_InsertItem (ctx->ctrl, &lvi) != pos)
	rc = -1;
    ctx->items++;
    return rc;
}


int
listview_add_item (listview_ctrl_t ctx, const char *text)
{
    int rc = 0;
    LV_ITEM lvi;

    memset (&lvi, 0, sizeof lvi);
    lvi.mask = LVIF_TEXT;
    lvi.pszText = (char *)text;
    rc = ListView_InsertItem (ctx->ctrl, &lvi);
    if( rc == -1 )
	rc = 1;	
    ctx->items++;
    return rc;
}


int
listview_add_item_image (listview_ctrl_t ctx, const char *text, int image)
{
    int rc = 0;
    LV_ITEM lvi;

    memset( &lvi, 0, sizeof lvi );
    lvi.mask = LVIF_TEXT | LVIF_IMAGE;
    lvi.pszText = (char *)text;
    lvi.iImage = image;
    rc = ListView_InsertItem (ctx->ctrl, &lvi);
    if (rc == -1)
	rc = 1;	
    ctx->items++;
    return rc;
}


int
listview_add_item2 (listview_ctrl_t ctx, const char *text, void *magic)
{
    int rc = 0;
    LVITEM lvi;

    memset (&lvi, 0, sizeof (lvi));
    lvi.mask = LVIF_TEXT | LVIF_PARAM;
    lvi.pszText = (char *)text;
    lvi.lParam = (LPARAM)magic;
    rc = ListView_InsertItem (ctx->ctrl, &lvi);
    if (rc == -1)
	rc = 1;	
    ctx->items++;
    return rc;
}


void*
listview_get_item2 (listview_ctrl_t ctx, int pos)
{
    LVITEM lvi;

    memset (&lvi, 0, sizeof lvi);
    lvi.mask = LVIF_PARAM;
    lvi.iItem = pos;
    ListView_GetItem (ctx->ctrl, &lvi);
    return (void*)lvi.lParam;
}


int
listview_set_item2 (listview_ctrl_t ctx, int pos, void *magic)
{
    LVITEM lvi;

    memset (&lvi, 0, sizeof (lvi));
    lvi.mask = LVIF_PARAM;
    lvi.iItem = pos;
    lvi.lParam = (LPARAM)magic;
    if (ListView_SetItem (ctx->ctrl, &lvi) == -1)
	return 1;
    return 0;
}


void
listview_add_sub_item (listview_ctrl_t ctx, int pos, int col, const char *text)
{
    ListView_SetItemText (ctx->ctrl, pos, col, (char*)text);
}


int
listview_count_items( listview_ctrl_t ctx, int curr_sel )
{	
    int n;

    n = curr_sel? ListView_GetSelectedCount( ctx->ctrl ) :    
     	          ListView_GetItemCount( ctx->ctrl );
    return n;
}


int
listview_del_column (listview_ctrl_t ctx, int pos)
{
    ctx->cols--;
    return ListView_DeleteColumn (ctx->ctrl, pos)? 0 : 1;
}


/* Delete a single item from @ctx at position @pos. */
int
listview_del_item (listview_ctrl_t ctx, int pos)
{
    int rc = 0;

    if (ListView_DeleteItem (ctx->ctrl, pos) == -1)
	rc = 1;
    return rc;
}


/* Delete all selected items in @ctx. */
int
listview_del_sel_items (listview_ctrl_t ctx)
{	
    int i, n;

    n = listview_count_items (ctx, 0);
    for (i = n;  i > -1; i--) {
	if (listview_get_item_state (ctx, i))
	    listview_del_item (ctx, i);
    }
    return 0;
}


int
listview_del_all_items (listview_ctrl_t ctx)
{	
    int rc = 0;

    if (ListView_DeleteAllItems (ctx->ctrl) == FALSE)
	rc = 1;
    return rc;
}


/* Return the index of the selected item.
   Support both selections and checkboxes. */
int
listview_get_selected_item (listview_ctrl_t lv)
{
    int pos, i;
    
    pos = listview_get_curr_pos (lv);
    if (pos != -1)
	return pos;
    for (i=0; i < listview_count_items (lv, 0); i++) {
	if (listview_get_item_state (lv, i))
	    return i;
    }
    return -1;
}


/* Return if the given element is selected or if the ext
   style is used, if the checkbox is activated. */
int
listview_get_item_state (listview_ctrl_t ctx, int pos)
{
    int ret;
    if (ctx->ext_chkbox)
	ret = ListView_GetCheckState (ctx->ctrl, pos);
    else
	ret = ListView_GetItemState (ctx->ctrl, pos, LVIS_SELECTED);
    return ret;
}


int
listview_sort_items (listview_ctrl_t ctx, int sortby, listview_cmp sort_cb)
{
    ListView_SortItems (ctx->ctrl, sort_cb, sortby);
    return 0;
}


int
listview_get_curr_pos (listview_ctrl_t ctx)
{	
    return ListView_GetNextItem( ctx->ctrl, -1, LVNI_SELECTED );
}


int
listview_get_item_text (listview_ctrl_t ctx, int pos, int col, char *text,
		        int maxbytes)
{
    if (pos == -1) {
	pos = listview_get_curr_pos (ctx);
	if (pos == -1)
	    return -1;
    }
    ListView_GetItemText (ctx->ctrl, pos, col, text, maxbytes);
    return 0;
}


/* Use extended style with checkboxes for each item. */
void
listview_set_chkbox_style (listview_ctrl_t ctx)
{
    ListView_SetExtendedListViewStyle (ctx->ctrl, LVS_EX_CHECKBOXES);
    ctx->ext_chkbox = 1;
}


/* Use extended style to select the entire row. */
void
listview_set_ext_style (listview_ctrl_t ctx)
{	
    ListView_SetExtendedListViewStyle (ctx->ctrl, LVS_EX_FULLROWSELECT);
}


int
listview_set_column_order( listview_ctrl_t ctx, int *array )
{	
    return ListView_SetColumnOrderArray( ctx->ctrl, ctx->cols, array );
}


void
listview_select_all (listview_ctrl_t ctx)
{	
    ListView_SetItemState( ctx->ctrl, -1, LVIS_SELECTED, LVIS_SELECTED );
}


void
listview_deselect_all (listview_ctrl_t ctx)
{
    ListView_SetItemState (ctx->ctrl, -1, ~LVNI_SELECTED, LVNI_SELECTED);
}


void
listview_select_one (listview_ctrl_t ctx, int pos)
{
    ListView_SetItemState (ctx->ctrl, pos, LVIS_SELECTED|LVIS_FOCUSED, 
					    LVIS_FOCUSED|LVIS_SELECTED);
}


void
listview_scroll (listview_ctrl_t ctx, int oldpos, int newpos)
{    
    const int size=14;
    
    if (oldpos == -1)
	oldpos = 0;
    //log_box ("debug", 0, "oldpos=%d newpos=%d diff=%d", 
		//oldpos, newpos, newpos-oldpos);
    ListView_Scroll (ctx->ctrl,	0, (newpos-oldpos)*size);
}


static int 
listview_find_substr (listview_ctrl_t ctx, const char *str)
{
    char buf[256];
    int i, n, fnd = 0;

    /* We assume the first column contains text. */
    n = listview_count_items (ctx, 0);
    for (i = 0; i < n; i++) {
	listview_get_item_text (ctx, i, 0, buf, DIM (buf)-1);
	if (stristr (buf, str)) {
	    fnd = 1;
	    break;
	}
    }
    if (!fnd)
	i = -1; /* not found */
    return i;
}


int
listview_find (listview_ctrl_t ctx, const char *str, int substr)
{
    LVFINDINFO inf;
    int pos;

    /* XXX: allow to make a substring search. */
    memset (&inf, 0, sizeof (inf));
    inf.flags = LVFI_STRING|LVFI_PARTIAL;
    inf.psz = str;
    pos = ListView_FindItem (ctx->ctrl, -1, &inf);
    if (pos == -1 && substr)
	pos = listview_find_substr (ctx, str);
    return pos;
}


void
listview_setview (listview_ctrl_t ctx, DWORD view)
{
    DWORD style = GetWindowLong (ctx->ctrl, GWL_STYLE);
    if ((style & LVS_TYPEMASK) != view)
	SetWindowLong (ctx->ctrl, GWL_STYLE, (style & ~LVS_TYPEMASK) | view);
}


void
listview_set_image_list (listview_ctrl_t ctx, int cx, int cy, 
			 HICON *ico, DWORD nicons)
{
    HIMAGELIST hil;

    if (cx <= 0 || cy <= 0)
	cx = cy = 16; /* default size. */

    hil = ImageList_Create (cx, cy, ILC_COLOR8|ILC_MASK, nicons, 1);
    ImageList_SetBkColor (hil, CLR_NONE);
    for (DWORD i=0; i < nicons; i++)
	ImageList_AddIcon (hil, ico[i]);    
    ListView_SetImageList (ctx->ctrl, hil, LVSIL_SMALL);    
}
